from typing import Dict, Optional

from smllib.sml import SmlBaseObj, SmlObjFieldInfo

from .sml_time import build_time, TIME_HINT


class SmlOpenResponse(SmlBaseObj):
    __sml__: Dict[str, SmlObjFieldInfo] = {
        'ref_time': SmlObjFieldInfo(func=build_time)
    }

    codepage: Optional[str]
    client_id: Optional[str]
    req_file_id: str
    server_id: str
    ref_time: TIME_HINT
    sml_version: Optional[int]


class SmlCloseResponse(SmlBaseObj):
    __sml__: Dict[str, SmlObjFieldInfo] = {}

    global_signature: Optional[str]



from typing import Dict, Optional, Tuple

from smllib.sml import SmlBaseObj, SmlListEntry, SmlObjFieldInfo

from .sml_time import build_time, TIME_HINT


class SmlGetListResponse(SmlBaseObj):
    __sml__: Dict[str, SmlObjFieldInfo] = {
        'act_sensor_time': SmlObjFieldInfo(func=build_time),
        'act_gateway_time': SmlObjFieldInfo(func=build_time),
    }

    client_id: Optional[str]
    server_id: str
    list_name: Optional[str]
    act_sensor_time: TIME_HINT
    val_list: Tuple[SmlListEntry, ...]
    list_signature: Optional[str]
    act_gateway_time: TIME_HINT



from typing import Dict, Type

import smllib.sml
from smllib.builder import SmlObjBuilder
from smllib.errors import EndOfSmlMsgExpected
from smllib.sml import EndOfSmlMsg
from smllib.sml._base_obj import SmlBaseObj
from smllib.sml_frame_snippet import SmlFrameSnippet


class SmlMessageBuilder(SmlObjBuilder[smllib.sml.SmlMessage]):
    BUILDS = smllib.sml.SmlMessage

    def build(self, obj: SmlFrameSnippet, classes: Dict[Type[SmlBaseObj], SmlObjBuilder]) -> smllib.sml.SmlMessage:

        lst = obj.get_value(list)
        if lst[5].value is not EndOfSmlMsg:
            raise EndOfSmlMsgExpected()
        lst.pop(5)

        return super().build(obj, classes)



from typing import Dict, Type, Union

from smllib.sml import SmlBaseObj, SmlChoice, SmlCloseResponse, SmlGetListResponse, SmlObjFieldInfo, SmlOpenResponse
from smllib.sml._base_obj import T_SML_OBJ as _T_SML_OBJ

MSG_TYPES: Dict[int, Type[_T_SML_OBJ]] = {
    # 0x0100: 'SmlOpenRequest',
    0x0101: SmlOpenResponse,
    # 0x0200: 'SmlCloseRequest',
    0x0201: SmlCloseResponse,
    # 0x0300: 'SmlGetProfilePackRequest',
    # 0x0301: 'SmlGetProfilePackResponse',
    # 0x0400: 'SmlGetProfileListRequest',
    # 0x0401: 'SmlGetProfileListResponse',
    # 0x0500: 'SmlGetProcParameterRequest',
    # 0x0501: 'SmlGetProcParameterResponse',
    # 0x0600: 'SmlSetProcParameterRequest',
    # 0x0601: 'SmlSetProcParameterResponse',
    # 0x0700: 'SmlGetListRequest',
    0x0701: SmlGetListResponse,
    # 0xff01: 'SmlAttentionResponse',
}


class SmlMessage(SmlBaseObj):
    __sml__: Dict[str, SmlObjFieldInfo] = {
        'message_body': SmlObjFieldInfo(choice=SmlChoice(MSG_TYPES))
    }

    transaction_id: str
    group_no: int
    abort_on_error: int
    message_body: Union[SmlOpenResponse, SmlCloseResponse, SmlGetListResponse]
    crc16: int



from typing import Final, TypeVar

INDENT: Final = '  '


class SmlBaseObj:
    __sml__: dict

    def format_msg(self, indent: int = 0):
        indent += 1
        r = f'<{self.__class__.__name__}>\n'
        w = max(map(len, self.__dict__), default=0)
        for k, v in self.__dict__.items():
            if isinstance(v, SmlBaseObj):
                r += f'{INDENT * indent}{str(k):s} {v.format_msg(indent)}'
            elif isinstance(v, (tuple, list)):
                r += f'{INDENT * indent}{str(k):s}:\n'
                for e in v:
                    r += f'{INDENT * (indent + 1)}{e.format_msg(indent + 1)}'
            else:
                r += f'{INDENT * indent}{str(k):{w}s}: {v}\n'
        return r


T_SML_OBJ = TypeVar('T_SML_OBJ', bound=SmlBaseObj, covariant=True)



from smllib.sml._base_obj import SmlBaseObj, T_SML_OBJ
from smllib.sml.sml_eom import EndOfSmlMsg

# isort: split

from smllib.sml.sml_choice import SmlChoice

# isort: split
from smllib.sml._field_info import inspect_obj, SmlObjFieldInfo

# isort: split

from smllib.sml.list_entry import SmlListEntry
from smllib.sml.response_get_list import SmlGetListResponse
from smllib.sml.response_open_close import SmlCloseResponse, SmlOpenResponse
from smllib.sml.sml_obis import ObisCode

# isort: split
from smllib.sml.message import SmlMessage



from binascii import a2b_hex
from typing import Dict, Type

from smllib.builder import SmlObjBuilder
from smllib.sml import SmlListEntry
from smllib.sml._base_obj import SmlBaseObj
from smllib.sml_frame_snippet import SmlFrameSnippet


class SmlListEntryBuilder(SmlObjBuilder[SmlListEntry]):
    BUILDS = SmlListEntry

    def build(self, obj: SmlFrameSnippet, classes: Dict[Type[SmlBaseObj], 'SmlObjBuilder']) -> SmlListEntry:
        ret = super().build(obj, classes)    # type: SmlListEntry

        value = ret.value
        if value is None:
            raise ValueError('value is required!')

        # Maybe it's ascii so we try to decode it
        if isinstance(value, str):
            v = a2b_hex(value).decode(errors='ignore')
            if v.isalnum():
                ret.value = v

        return ret



from typing import Dict, Type

from smllib.sml import SmlCloseResponse, SmlGetListResponse, SmlListEntry, SmlMessage, SmlOpenResponse, T_SML_OBJ

from . import SmlCloseResponseBuilder, SmlGetListResponseBuilder, \
    SmlListEntryBuilder, SmlMessageBuilder, SmlOpenResponseBuilder, T_SML_BUILDER

CTX_HINT = Dict[Type[T_SML_OBJ], T_SML_BUILDER]


def create_context() -> CTX_HINT:
    return {
        SmlListEntry: SmlListEntryBuilder(),
        SmlGetListResponse: SmlGetListResponseBuilder(),
        SmlMessage: SmlMessageBuilder(),
        SmlCloseResponse: SmlCloseResponseBuilder(),
        SmlOpenResponse: SmlOpenResponseBuilder()
    }



from typing import Optional

from smllib.builder import create_context, CTX_HINT
from smllib.crc import get_crc
from smllib.errors import CrcError
from smllib.sml_frame import SmlFrame


class SmlStreamReader:
    MAX_SIZE = 50 * 1024

    def __init__(self, build_ctx: CTX_HINT = None):
        self.bytes: bytes = b''
        self.build_ctx: CTX_HINT = build_ctx if build_ctx is not None else create_context()

    def add(self, _bytes: bytes):
        self.bytes += _bytes
        if len(self.bytes) > SmlStreamReader.MAX_SIZE:
            self.bytes = self.bytes[-1 * SmlStreamReader.MAX_SIZE:]

    def clear(self):
        self.bytes = b''

    def get_frame(self) -> Optional[SmlFrame]:
        start = self.bytes.find(b'\x1B\x1B\x1B\x1B\x01\x01\x01\x01')
        if start == -1:
            return None

        # if we start reading in the mid of a message
        if start != 0:
            self.bytes = self.bytes[start:]
            start = 0

        end = -1
        while (end := self.bytes.find(b'\x1B\x1B\x1B\x1B\x1A', end + 1)) != -1:
            pre = self.bytes[end - 4: end]
            if pre != b'\x1B\x1B\x1B\x1B':
                break

        if end == -1:
            return None

        end += 8
        if len(self.bytes) < end:
            return None

        # remove msg from buffer
        msg = self.bytes[start:end]
        self.bytes = self.bytes[end:]

        # Last three bytes are PADDING, CRC PART 1, CRC PART 2
        padding = msg[-3]

        # check crc
        crc_msg = msg[-2] << 8 | msg[-1]
        crc_calc = get_crc(msg[:-2])
        if crc_msg != crc_calc:
            raise CrcError(msg, crc_msg, crc_calc)

        frame = msg[8: -1 * (8 + padding)].replace(b'\x1B\x1B\x1B\x1B\x1B\x1B\x1B\x1B', b'\x1B\x1B\x1B\x1B')
        return SmlFrame(frame, msg_ctx=msg, build_ctx=self.build_ctx)



from typing import List, Optional

from smllib.builder import create_context, CTX_HINT
from smllib.errors import InvalidBufferPos
from smllib.sml import EndOfSmlMsg, SmlListEntry, SmlMessage
from smllib.sml_frame_snippet import SmlFrameSnippet


class SmlFrame:
    def __init__(self, buffer: bytes, build_ctx: CTX_HINT = None, msg_ctx: Optional[bytes] = None):
        self.bytes = buffer
        self.buffer = memoryview(buffer)
        self.buf_len = len(buffer)

        self.next_pos = 0

        self.msg_ctx = msg_ctx  # This is the whole sml message
        self.build_ctx: CTX_HINT = build_ctx if build_ctx is not None else create_context()

    def get_value(self, pos: Optional[int] = None) -> SmlFrameSnippet:
        if pos is None:
            pos = self.next_pos
        snip_start = pos

        # check start pos
        if pos >= self.buf_len:
            raise InvalidBufferPos(f'Start pos bigger than buffer: {pos} > {self.buf_len}')

        # advance
        v = self.buffer[pos]
        start = pos + 1

        # ----------------------------------------
        # types with fixed size
        # ----------------------------------------

        # No value
        if v == 0x01:
            self.next_pos = start
            return SmlFrameSnippet(None, snip_start, start, self.buffer)

        # End of a SmlMSg
        if v == 0x00:
            self.next_pos = start
            return SmlFrameSnippet(EndOfSmlMsg, snip_start, start, self.buffer)

        # Bool
        if v == 0x42:
            self.next_pos = start + 1
            return SmlFrameSnippet(bool(self.buffer[start]), snip_start, self.next_pos, self.buffer)

        # ----------------------------------------
        # types with dynamic size
        # ----------------------------------------
        is_long = bool(v & 0x80)
        _type = v & 0x70  # type
        _size = v & 0x0F  # size including the 1-byte tag
        while is_long:
            v = self.buffer[start]
            _size = _size << 4 | v & 0x0F
            is_long = bool(v & 0x80)
            start += 1

        # type is a list
        if _type == 0x70:
            self.next_pos = start   # Must be s_pos because we can have lists with a long length
            return SmlFrameSnippet([None for _ in range(_size)], snip_start)

        # End position
        end = pos + _size
        if end > self.buf_len:
            raise InvalidBufferPos(f'Pos bigger than buffer: {end} > {self.buf_len}')
        self.next_pos = end

        # 0x50: signed integer, 0x60 unsigned integer
        if _type == 0x50 or _type == 0x60:
            return SmlFrameSnippet(
                int.from_bytes(self.buffer[start:end], byteorder='big', signed=_type == 0x50),
                snip_start, end, self.buffer
            )

        # 0x00: octet str
        if _type == 0x00:
            return SmlFrameSnippet(self.buffer[start:end].hex(), snip_start, end, self.buffer)

        raise ValueError(f'Unknown data type: {_type:02x}!')

    def parse_frame(self) -> List[SmlMessage]:
        ret = []
        self.next_pos = 0
        while self.next_pos < self.buf_len:

            if not self.buffer[self.next_pos] == 0x76:
                raise ValueError(
                    f'No start of SML Message found at {self.next_pos}: 0x{self.buffer[self.next_pos]:x}\n'
                    f'{self.buffer.hex()}'
                )

            # This will always return a list
            val = self._parse_msg(self.get_value())
            ret.append(self.build_ctx[SmlMessage].build(val, self.build_ctx))
        return ret

    def _parse_msg(self, parent_obj: SmlFrameSnippet) -> SmlFrameSnippet:
        # it's always a list now
        _lst = parent_obj.get_value(list)
        for i, _ in enumerate(_lst):
            _lst[i] = v = self.get_value()
            if isinstance(v.value, list):
                self._parse_msg(v)
        parent_obj.stop_pos(self.next_pos, self.buffer)
        return parent_obj

    def get_obis(self) -> List[SmlListEntry]:
        """Returns all obis values in the frame without parsing the frame"""
        ret = []
        pos = 0
        while (pos := self.bytes.find(b'\x77\x07\x01', pos)) != -1:
            data = self.get_value(pos)
            if not isinstance(data.value, list):
                continue

            self._parse_msg(data)
            ret.append(self.build_ctx[SmlListEntry].build(data, self.build_ctx))

            # Don't search in the frame again since the payload might contain '770701'
            pos = self.next_pos
        return ret



from typing import Optional, Type, TypeVar, Union

from smllib.errors import WrongValueType
from smllib.sml.sml_eom import CEndOfSmlMsg

SNIP_TYPE = TypeVar('SNIP_TYPE', bound=object)


class SmlFrameSnippet:
    __slots__ = ('pos', 'value', 'msg')

    def __init__(self, value: Union[None, bool, int, str, float, list, CEndOfSmlMsg], start: int,
                 stop: Optional[int] = None, buf: Optional[memoryview] = None):

        msg = None
        if stop is not None:
            if buf is None:
                raise ValueError('Arg stop und buf must be used together')
            msg = buf[start: stop]

        self.value = value
        self.pos = start
        self.msg: Optional[memoryview] = msg

    def stop_pos(self, pos: int, buf: memoryview) -> 'SmlFrameSnippet':
        assert self.msg is None
        self.msg = buf[self.pos: pos]
        return self

    def get_value(self, val_type: Type[SNIP_TYPE]) -> SNIP_TYPE:
        value = self.value
        if not isinstance(value, val_type):
            raise WrongValueType(f'Expected type {val_type.__name__} but got type {type(value).__name__}')
        return value



from smllib.builder import SmlObjBuilder
from smllib.sml import SmlCloseResponse, SmlOpenResponse


class SmlOpenResponseBuilder(SmlObjBuilder[SmlOpenResponse]):
    BUILDS = SmlOpenResponse


class SmlCloseResponseBuilder(SmlObjBuilder[SmlCloseResponse]):
    BUILDS = SmlCloseResponse



from typing import Any, List, Mapping, Tuple, Type, Union

from smllib.errors import UnsupportedChoiceValue, WrongArgCount
from smllib.sml import T_SML_OBJ
from smllib.sml_frame_snippet import SmlFrameSnippet


class SmlChoice:
    def __init__(self, choices: Mapping[int, Type[T_SML_OBJ]]):
        self.choices = choices

    def get(self, obj: List[SmlFrameSnippet]) -> Union[Tuple[None, Any], Tuple[Type[T_SML_OBJ], Any]]:
        if len(obj) != 2:
            raise WrongArgCount()

        _type, _value = obj  # type: SmlFrameSnippet, SmlFrameSnippet
        choice_value = _type.get_value(int)

        ret = self.choices.get(choice_value)
        if ret is None:
            raise UnsupportedChoiceValue(choice_value)

        return ret, _value

    def __eq__(self, other):
        return self.choices == other.choices



from ._builder import SmlObjBuilder, T_SML_BUILDER

# isort: split

from .list_entry import SmlListEntryBuilder
from .message import SmlMessageBuilder
from .response_get_list import SmlGetListResponseBuilder
from .response_open_close import SmlCloseResponseBuilder, SmlOpenResponseBuilder

# isort: split

from .default_context import create_context, CTX_HINT



from datetime import datetime, timedelta
from typing import Union

from smllib.errors import UnsupportedChoiceValue

TIME_HINT = Union[None, int, datetime]


def build_time(_in):
    if _in is None:
        return _in

    # This is a workaround for times that are not reported according to specification
    # Instead of a choice list these devices report just the timestamp - however I am unsure about it.
    if isinstance(_in, int):
        return _in

    type_s, value_s = _in
    _type = type_s.value
    if _type == 1:
        return value_s.value
    if _type == 2:
        return datetime.utcfromtimestamp(value_s.value)
    if _type == 3:
        ts, offset1, offset2 = value_s.value
        return datetime.utcfromtimestamp(ts.value) + timedelta(minutes=offset1.value) + timedelta(minutes=offset2.value)

    raise UnsupportedChoiceValue(_type)



from typing import Dict, Generic, Tuple, Type, TypeVar

from smllib.errors import WrongArgCount, WrongValueType
from smllib.sml import inspect_obj, SmlBaseObj, SmlObjFieldInfo, T_SML_OBJ
from smllib.sml_frame_snippet import SmlFrameSnippet


class SmlObjBuilder(Generic[T_SML_OBJ]):
    BUILDS: Type[T_SML_OBJ]

    def __init__(self):
        assert issubclass(self.BUILDS, SmlBaseObj), self.BUILDS
        self.fields: Dict[str, SmlObjFieldInfo] = inspect_obj(self.BUILDS)

    def build(self, obj: SmlFrameSnippet, classes: Dict[Type[SmlBaseObj], 'SmlObjBuilder']) -> T_SML_OBJ:
        # check length
        lst = obj.get_value(list)
        if len(self.fields) != len(lst):
            raise WrongArgCount()

        out = self.BUILDS()
        for i, a in enumerate(self.fields.items()):   # type: int, Tuple[str, SmlObjFieldInfo]
            name, field = a
            value = lst[i].value

            # rebuild with choice
            if field.choice is not None:
                choice_cls, value = field.choice.get(value)
                value = classes[choice_cls].build(value, classes)

            func = field.func
            if func is not None:
                value = func(value)

            if field.is_container:
                cls_builder = classes[field.type]
                value = tuple([cls_builder.build(v, classes) for v in value])
            else:
                if not isinstance(value, field.type):
                    raise WrongValueType(f'{value} ({type(value)}) != {field.type}')

            setattr(out, name, value)
        return out


T_SML_BUILDER = TypeVar('T_SML_BUILDER', bound=SmlObjBuilder, covariant=True)



from smllib import errors
from smllib.__version__ import __version__

# isort: split

import smllib.reader
import smllib.sml
from smllib.reader import SmlStreamReader
from smllib.sml_frame import SmlFrame



import typing


class SmlLibException(Exception):
    pass


class CrcError(SmlLibException):
    def __init__(self, msg: bytes, crc_msg: int, crc_calc: int):
        self.msg = msg
        self.crc_msg = crc_msg
        self.crc_calc = crc_calc

    def __repr__(self):
        return f'<{self.__class__.__name__} msg: {self.crc_msg:04x} calc: {self.crc_calc:04x}>'


class InvalidBufferPos(SmlLibException):
    pass


class EndOfSmlMsgExpected(SmlLibException):
    pass


class WrongArgCount(SmlLibException):
    pass


class WrongValueType(SmlLibException):
    pass


class FieldValueTypeDoesNotMatch(SmlLibException):
    def __init__(self, cls, field: str, type: typing.Type[typing.Union[str, int, float]], *args: object) -> None:
        super().__init__(*args)
        self.cls = cls
        self.field: str = field
        self.type: typing.Type[typing.Union[str, int, float]] = type

    def __str__(self) -> str:
        return f'Field value type does not match for for {self.field} in {self.cls.__name__}: {self.type.__name__}'


class UnsupportedChoiceValue(SmlLibException):
    def __init__(self, type: int, *args: object) -> None:
        super().__init__(*args)
        self.type: int = type

    def __str__(self) -> str:
        width = 2 if self.type <= 255 else 4
        return f'Unsupported type 0x{self.type:0{width}x}'



class ObisCode(str):
    def __init__(self, obis_hex: str):
        _a = int(obis_hex[0:2], 16)
        _b = int(obis_hex[2:4], 16)
        _c = int(obis_hex[4:6], 16)
        _d = int(obis_hex[6:8], 16)
        _e = int(obis_hex[8:10], 16)
        _f = int(obis_hex[10:12], 16)
        self.obis_code = f'{_a}-{_b}:{_c}.{_d}.{_e}*{_f}'
        self.obis_short = f'{_c}.{_d}.{_e}'

    def __new__(cls, obis_hex: str):
        obj = str.__new__(cls, obis_hex)
        return obj


def build_obis(_in) -> ObisCode:
    if not isinstance(_in, str) or len(_in) != 12:
        raise ValueError(f'Obis code must be a 12 char hex str (is "{_in}" {type(_in)})!')

    return ObisCode(_in)



import typing
from pathlib import Path

import setuptools  # type: ignore


def load_version() -> str:
    version: typing.Dict[str, str] = {}
    with open("src/smllib/__version__.py") as fp:
        exec(fp.read(), version)
    assert version['__version__'], version
    return version['__version__']


__version__ = load_version()

print(f'Version: {__version__}')
print('')

# When we run tox tests we don't have these files available so we skip them
readme = Path(__file__).with_name('readme.md')
long_description = ''
if readme.is_file():
    with readme.open("r", encoding='utf-8') as fh:
        long_description = fh.read()

setuptools.setup(
    name="smllib",
    version=__version__,
    author="spaceman_spiff",
    # author_email="",
    description="A library for the SML (Smart Message Language) protocol",
    keywords=[
        'sml',
        'obis',
        'smart message language',
        'energy meter',
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/spacemanspiff2007/SmlLib",
    project_urls={
        'GitHub': "https://github.com/spacemanspiff2007/SmlLib",
    },
    packages=setuptools.find_packages(where='src', exclude=['tests*']),
    package_dir={'': 'src'},
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Home Automation"
    ],
)



from typing import Dict, Optional, Union

from smllib.const import OBIS_NAMES, UNITS
from smllib.sml import SmlObjFieldInfo
from smllib.sml._base_obj import INDENT, SmlBaseObj

from .sml_obis import build_obis, ObisCode
from .sml_time import build_time, TIME_HINT


class SmlListEntry(SmlBaseObj):
    __sml__: Dict[str, SmlObjFieldInfo] = {
        'val_time': SmlObjFieldInfo(func=build_time),
        'obis': SmlObjFieldInfo(func=build_obis)
    }

    obis: ObisCode
    status: Optional[int]
    val_time: TIME_HINT
    unit: Optional[int]
    scaler: Optional[int]
    value: Union[None, str, int, float]
    value_signature: Optional[str]

    def __repr__(self):
        r = []
        for k, v in self.__dict__.items():
            if v is not None:
                r.append(f'{k}: {v}')
        return f'<{", ".join(r)}>'

    def format_msg(self, indent: int = 0):
        indent += 1
        r = f'<{self.__class__.__name__}>\n'
        w = max(map(len, self.__dict__), default=0)

        for k, v in self.__dict__.items():
            r += f'{INDENT*indent}{str(k):{w}s}: {v}{f" ({self.obis.obis_code})" if k == "obis" else ""}\n'

        summary = ''
        if self.unit:
            val = self.get_value()
            u = UNITS.get(self.unit)
            if u is None:
                u = f" ?:{self.unit}"
            summary += f'{val}{u}'

        desc = OBIS_NAMES.get(self.obis)
        if desc is not None:
            summary += f'{" " if summary else ""}({desc})'
        if summary:
            r += f'{INDENT*indent}-> {summary:s}\n'
        return r

    def get_value(self) -> Union[None, float, str]:
        if self.value is None:
            return None

        # Some devices report the texts with scaler 0
        if self.scaler is None or isinstance(self.value, str):
            return self.value

        return round(self.value * 10**self.scaler, abs(self.scaler) + 3)



from typing import Any, Callable, Dict, Final, get_args, get_origin, get_type_hints, Optional, Type, Union

from smllib.sml import SmlBaseObj, SmlChoice


class SmlObjFieldInfo:
    def __init__(self, func: Optional[Callable[[Any], Any]] = None, type=None,
                 choice: Optional[SmlChoice] = None, is_container=False):
        self.func: Final = func
        self.choice: Final = choice
        self.type = type
        self.is_container = is_container

    def __eq__(self, other):
        return self.func == other.func and self.type == other.type and \
            self.choice == other.choice and self.is_container == other.is_container

    def copy(self) -> 'SmlObjFieldInfo':
        return self.__class__(func=self.func, type=self.type, choice=self.choice, is_container=self.is_container)

    def __repr__(self):
        c = []
        if self.func is not None:
            c.append(f'func={self.func.__name__}')
        if self.type is not None:
            c.append(f'type={self.type.__name__}')
        if self.choice is not None:
            c.append(f'choice={self.choice}')
        if self.is_container is not None:
            c.append(f'is_container={self.is_container}')
        return f'<{self.__class__.__name__} {", ".join(c)} at {id(self):X}>'


def inspect_obj(cls: Type['SmlBaseObj']) -> Dict[str, SmlObjFieldInfo]:
    fields: Dict[str, SmlObjFieldInfo] = {}

    for name, hint in get_type_hints(cls).items():
        if name == '__sml__':
            continue

        info = cls.__sml__.get(name)
        if info is None:
            info = SmlObjFieldInfo()
        else:
            assert isinstance(info, SmlObjFieldInfo)
            info = info.copy()
        fields[name] = info

        origin = get_origin(hint)
        if origin is None:
            info.type = hint
            continue

        # Union[str, int]
        if origin is Union:
            info.type = get_args(hint)
            continue

        # Tuple[type, ...]
        if origin is tuple:
            info.is_container = True
            args = get_args(hint)
            assert args[1] is ...
            info.type = args[0]
            continue

        raise ValueError(f'Unknown hint: {origin} for {name}')

    return fields
